/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#include "aidge/quantization/PTQ/CLE.hpp"

#include <cmath>    // std::abs, std::fabs, std::sqrt
#include <cstddef>  // std::size_t
#include <memory>
#include <vector>

#include "aidge/quantization/PTQ/Clipping.hpp"
#include "aidge/quantization/PTQ/PTQ.hpp"  // retrieveNodeVector

#include "aidge/graph/GraphView.hpp"

#include "aidge/scheduler/SequentialScheduler.hpp"
#include "aidge/scheduler/Scheduler.hpp"
#include "aidge/utils/Log.hpp"
#include "aidge/operator/OperatorTensor.hpp"
#include "aidge/utils/Log.hpp"

#include "aidge/operator/Mul.hpp"
#include "aidge/operator/ArgMax.hpp"
#include "aidge/operator/Abs.hpp"
#include "aidge/operator/Reshape.hpp"
#include "aidge/operator/Round.hpp"

#include "aidge/operator/Mul.hpp"
#include "aidge/operator/ArgMax.hpp"
#include "aidge/operator/Abs.hpp"
#include "aidge/operator/Reshape.hpp"
#include "aidge/operator/Round.hpp"

namespace Aidge
{

static std::shared_ptr<Tensor> getWeightTensor(std::shared_ptr<Node> node)
{
    return std::static_pointer_cast<OperatorTensor>(node->getOperator())->getInput(1);
}

static std::shared_ptr<Tensor> getBiasTensor(std::shared_ptr<Node> node)
{
    return std::static_pointer_cast<OperatorTensor>(node->getOperator())->getInput(2);
}

static void rescaleTensor(std::shared_ptr<Tensor> tensor, double scaling)
{
    auto mulOp = Mul_Op();
    mulOp.setDataType(tensor->dataType());
    mulOp.setBackend(tensor->backend());

    std::shared_ptr<Aidge::Tensor> scalingTensor = std::make_shared<Aidge::Tensor>(Aidge::Array1D<double, 1> {scaling});
    scalingTensor->setDataType(tensor->dataType());
    scalingTensor->setBackend(tensor->backend());

    mulOp.associateInput(0, tensor);
    mulOp.associateInput(1, scalingTensor);

    mulOp.forward();
    
    auto outTensor = mulOp.getOutput(0);
    *tensor = *outTensor;
    //tensor->copyCast(*outTensor);
}

// TODO : make the retreival of argmax values backend independant (refCastFrom)
static double getTensorAbsoluteMax(std::shared_ptr<Tensor> tensor)
{
    // get the abs tensor
    std::shared_ptr<Tensor> fallback; //Fallback tensor for refCastFR
    std::shared_ptr<Tensor> absTensor = std::make_shared<Tensor>(tensor->abs());

    // flatten the abs tensor

    std::int64_t nbElement = tensor->size();

    auto reshapeOp = Reshape_Op({nbElement});
    reshapeOp.setDataType(tensor->dataType());
    reshapeOp.setBackend(tensor->backend());

    reshapeOp.associateInput(0, absTensor);
    reshapeOp.forward();
    std::shared_ptr<Tensor> flatTensor = reshapeOp.getOutput(0);
    const Tensor& localFlatTensor = flatTensor->refCastFrom(fallback, DataType::Float64, "cpu"); 

    // Get the argmax

    auto argmaxOp = ArgMax_Op(0, true, false);
    argmaxOp.setDataType(tensor->dataType());
    argmaxOp.setBackend(tensor->backend());

    argmaxOp.associateInput(0, flatTensor);
    argmaxOp.forward();

    const Tensor& argMaxTensor = argmaxOp.getOutput(0)->refCastFrom(fallback, DataType::Float64, "cpu"); 

    // Return the max

    int maxIndex = std::round(argMaxTensor.get<double>(0));

    return localFlatTensor.get<double>(maxIndex);
}
//Function used to extraxt the local tensor (from a ProducerScalingNode)
std::shared_ptr<Aidge::Tensor> getLocalTensor(std::shared_ptr<Node> node) {
    if (node->getParent(1)->attributes()->hasAttr("quantization.ptq.isProducerScaling")) {
        std::shared_ptr<Aidge::OperatorTensor> operatorTensor = std::static_pointer_cast<OperatorTensor>(node->getParent(1)->getOperator());
        operatorTensor->forward();// We need the forward pass to compute the scaled value of the Tensor
        return operatorTensor->getOutput(0);
    } else {
        return getWeightTensor(node);
    }
}

void crossLayerEqualization(std::shared_ptr<GraphView> graphView, double targetDelta)
{
    std::vector<std::shared_ptr<Node>> nodeVector = retrieveNodeVector(graphView);

    // Check if the CLE can be applied ...

    for (std::shared_ptr<Node> node : nodeVector)
    {
        if (node->getChildren().size() > 1) {
            Log::notice(" Network have multiple branches, skipping the CLE ... ");
            return;
        }
        if (isNotQuantized(node)) {
            Log::notice(" Network contains non linear nodes, skipping the CLE ... ");
            return;
        }
    }

    Log::info(" Applying the Cross-Layer Equalization ... ");

    // Get the vector of affine nodes

    std::vector<std::shared_ptr<Node>> affineNodeVector;
    for (std::shared_ptr<Node> node : nodeVector)
        if (isAffine(node))
            affineNodeVector.push_back(node);

    double maxRangeDelta;

    do
    {
        maxRangeDelta = 0.0;
        
        for (size_t i = 0; i < (affineNodeVector.size() - 1); i++)
        {
            std::shared_ptr<Node> n1 = affineNodeVector[i];
            std::shared_ptr<Node> n2 = affineNodeVector[i+1];

            std::shared_ptr<Aidge::Tensor> n1localTensor = getLocalTensor(n1);
            std::shared_ptr<Aidge::Tensor> n2localTensor = getLocalTensor(n2);
            
            double r1 = getTensorAbsoluteMax(n1localTensor);
            double r2 = getTensorAbsoluteMax(n2localTensor);

            double s1 = std::sqrt(r1 * r2) / r1;
            double s2 = std::sqrt(r1 * r2) / r2;

            insertScalingBelowProducer(n1->getParent(1), s1, graphView);
            insertScalingBelowProducer(n2->getParent(1), s2, graphView);
            insertScalingBelowProducer(n1->getParent(2), s1, graphView);

            double rangeDelta = std::abs(r1 - r2);
            if (rangeDelta > maxRangeDelta)
                maxRangeDelta = rangeDelta;
        }
    }
    while (maxRangeDelta > targetDelta);
}

}