/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#include "aidge/quantization/PTQ/CLE.hpp"
#include "aidge/quantization/PTQ/Clipping.hpp"
#include "aidge/quantization/PTQ/PTQ.hpp"
#include "aidge/operator/PTQMetaOps.hpp"

#include "aidge/data/Tensor.hpp"
#include "aidge/graph/GraphView.hpp"
#include "aidge/graph/Node.hpp"
#include "aidge/scheduler/SequentialScheduler.hpp"
#include "aidge/scheduler/Scheduler.hpp"
#include "aidge/utils/Log.hpp"

#include "aidge/operator/Producer.hpp"
#include "aidge/operator/Mul.hpp"
#include "aidge/operator/Round.hpp"
#include "aidge/operator/ReLU.hpp"
#include "aidge/operator/BatchNorm.hpp"
#include "aidge/operator/Conv.hpp"
#include "aidge/operator/ArgMax.hpp"
#include "aidge/operator/Reshape.hpp"

#include "aidge/recipes/Recipes.hpp"
#include "aidge/recipes/QuantRecipes.hpp"

namespace Aidge
{

bool isAffine(std::shared_ptr<Node> node)
{
    return (affineNodeTypes.find(node->type()) != affineNodeTypes.end());
}

bool isSeamless(std::shared_ptr<Node> node)
{
    return (seamlessNodeTypes.find(node->type()) != seamlessNodeTypes.end());
}

bool isMerging(std::shared_ptr<Node> node)
{
    return (mergingNodeTypes.find(node->type()) != mergingNodeTypes.end());
}

bool isNotQuantized(std::shared_ptr<Node> node)
{
    return (notQuantizedNodeTypes.find(node->type()) != notQuantizedNodeTypes.end());
}

bool checkArchitecture(std::shared_ptr<GraphView> graphView)
{
    std::set<std::string> otherNodeTypes({"Flatten", "Softmax", "BatchNorm2D", "ReLU", "Producer"});

    std::set<std::string> notQuantizedNodesTypes;

    for (std::shared_ptr<Node> node : graphView->getNodes())
    {
        bool isOther = otherNodeTypes.find(node->type()) != otherNodeTypes.end();
        if (!isOther && !isAffine(node) && !isSeamless(node) && !isMerging(node) && !isNotQuantized(node)) {
            Log::warn(" GraphView can't be quantized : node type {} is not supported !", node->type());
            return false;
        }

        if (isNotQuantized(node))
            notQuantizedNodesTypes.insert(node->type());
    }

    if (!notQuantizedNodesTypes.empty()) {
        std::string tokens;
        for (std::string s : notQuantizedNodesTypes)
            tokens += (s + " ");
        Log::warn(" Network contains non-linear nodes that won't be quantized : {}", tokens);
    }

    return true;
}

void prepareNetwork(std::shared_ptr<GraphView> graphView)
{
    removeFlatten(graphView);
    sanitizeNodeNames(graphView);

    bool containsBatchNorm = false;
    std::vector<std::shared_ptr<Node>> nodeVector = retrieveNodeVector(graphView);

    for (std::shared_ptr<Node> node : nodeVector)
        if (node->type() == "BatchNorm") {
            containsBatchNorm = true;
            break;
        }

    if (containsBatchNorm)
        fuseBatchNorm(graphView);   

    popSoftMax(graphView);
}

static std::shared_ptr<Aidge::Node> getUniqueChildren(std::shared_ptr<Aidge::Node> node) 
{
    std::set<std::shared_ptr<Aidge::Node>> childrenSet = node->getChildren();
    AIDGE_ASSERT(childrenSet.size() == 1, " Attempted to access to a unique child while the parent have multiple ones ! ");
    return *(childrenSet.begin());
}

static std::string determineBackend(std::shared_ptr<Aidge::Node> node)
{
    std::string backend = node->getOperator()->backend();

    if (backend != "")
        return backend;
    else 
    {
        // gather the parent backends

        std::set<std::string> parentBackends;
        for (auto parent : node->getParents())
            parentBackends.insert(determineBackend(parent)); // it always answers a non empty value !

        // check if we have two or more different backends gathered

        if (parentBackends.size() > 1) 
        {
            Log::warn(" Unable to determine backend of node {} due to conflicting parent ones !", node->name());
            return (*parentBackends.begin());
        }
        
        // if all parents have the same backend return it

        if (parentBackends.size() == 1) 
            return (*parentBackends.begin());
    }

    return "cpu"; // escape path when no parents are found
}

static int getInputIndex(std::shared_ptr<Node> node, std::shared_ptr<Node> parentNode)
{
    int index = 0;
    while (node->getParent(index) != parentNode) 
        index++;
    return index;
}

void multiplyScalingFactor(std::shared_ptr<Aidge::Node> node, double coeff)
{
    AIDGE_ASSERT(node->type() == "Mul" && (node->attributes()->hasAttr("quantization.ptq.isProducerScaling") || node->attributes()->hasAttr("quantization.ptq.isScaling")),
    "Cannot update the scaling factor on Node of type {} with no scaling tag", node->type());
    
    auto scalingFactorTensor = std::static_pointer_cast<OperatorTensor>(node->getOperator())->getInput(1);

    std::shared_ptr<Tensor> fallback;
    const Tensor& localTensor = scalingFactorTensor->refCastFrom(fallback, DataType::Float64, "cpu");
    double previousScalingFactor = localTensor.get<double>(0);

    std::shared_ptr<Tensor> resultTensor = std::make_shared<Tensor>(Array1D<double, 1> {previousScalingFactor * coeff});
    node->input(1).first->getOperator()->setOutput(0, resultTensor);
}

// Utility function that insert a node below another one already connected 
static void insertChildren(std::shared_ptr<Node> parent, std::shared_ptr<Node> newNode, std::shared_ptr<GraphView> graphView) 
{
    // Checking the parents always have at least 1 children

    AIDGE_ASSERT(parent->getChildren().size() > 0, " Parent node must have at least one child to insert a new node ! ");

    // Retreive children connection indexes

    std::vector<std::shared_ptr<Node>> nextNodes = parent->getChildren(0);
    std::vector<int> inputIndices(nextNodes.size());
    for (std::size_t i = 0; i < nextNodes.size(); i++) {
        inputIndices[i] = getInputIndex(nextNodes[i], parent);
    }

    // Disconnect childs from parent

    for (std::shared_ptr<Node> nextNode : nextNodes) {
        parent->removeChild(nextNode, 0);
    }

    // Insert the new node between the child and the parent

    parent->addChild(newNode, 0, 0);
    for (std::size_t i = 0; i < nextNodes.size(); i++) {
        newNode->addChild(nextNodes[i], 0, inputIndices[i]);
    }

    graphView->add(newNode);
}

bool insertRoundBelowProducer(std::shared_ptr<Node> node, std::shared_ptr<GraphView> graphView)
{
    if (node->attributes()->hasAttr("quantization.ptq.isProducerScaling") && node->type() != "Round")
    {
        std::shared_ptr<Aidge::Node> roundNode = Round(node->name() + "_Round");
        roundNode->getOperator()->setDataType(DataType::Float64); // getDataType(parentNode)
        roundNode->getOperator()->setBackend(determineBackend(node));

        insertChildren(node, roundNode, graphView);
        roundNode->attributes()->addAttr("quantization.ptq.isProducerRounding", 0.0);
        return true;
    }
    return false;
}

static double getTensorAbsoluteMax(std::shared_ptr<Tensor> tensor)
{
    // get the abs tensor
    std::shared_ptr<Tensor> fallback; //Fallback tensor for refCastFR

    std::shared_ptr<Tensor> absTensor = std::make_shared<Tensor>(tensor->abs());

    // flatten the abs tensor

    std::int64_t nbElement = tensor->size();

    auto reshapeOp = Reshape_Op({nbElement});
    reshapeOp.setDataType(tensor->dataType());
    reshapeOp.setBackend(tensor->backend());

    reshapeOp.associateInput(0, absTensor);
    reshapeOp.forward();
    std::shared_ptr<Tensor> flatTensor = reshapeOp.getOutput(0);
    const Tensor& localFlatTensor = flatTensor->refCastFrom(fallback, DataType::Float64, "cpu"); 

    // Get the argmax

    auto argmaxOp = ArgMax_Op(0, true, false);
    argmaxOp.setDataType(tensor->dataType());
    argmaxOp.setBackend(tensor->backend());

    argmaxOp.associateInput(0, flatTensor);
    argmaxOp.forward();
    const Tensor& argMaxTensor = argmaxOp.getOutput(0)->refCastFrom(fallback, DataType::Float64, "cpu"); 

    // Return the max

    int maxIndex = std::round(argMaxTensor.get<double>(0));

    return localFlatTensor.get<double>(maxIndex);
}


// TODO : pass nodeVector by reference ...
static std::vector<std::shared_ptr<Node>> removeMatchingNodes(std::vector<std::shared_ptr<Node>> nodeVector, std::string nodeType)
{
    std::vector<std::shared_ptr<Node>> remainingNodes;
    for (std::shared_ptr<Node> node : nodeVector)
        if (node->type() != nodeType)
            remainingNodes.push_back(node);

    return remainingNodes;
}

static std::vector<std::shared_ptr<Node>> removeProdScalingNodes(std::vector<std::shared_ptr<Node>> nodeVector)
{
    std::vector<std::shared_ptr<Node>> remainingNodes;
    for (std::shared_ptr<Node> node : nodeVector)
        if (!node->attributes()->hasAttr("quantization.ptq.isProducerScaling"))
            remainingNodes.push_back(node);

    return remainingNodes;
}

static void fixScheduling(std::vector<std::shared_ptr<Node>>& nodeVector) {

    std::vector<std::shared_ptr<Node>> correctedVector;

    for (int i = (nodeVector.size() - 1); i >= 0; --i)
    {
        std::shared_ptr<Node> node = nodeVector[i];
        bool isAlreadyInside = (std::find(correctedVector.begin(), correctedVector.end(), node) != correctedVector.end());
        if (!isAlreadyInside)
            correctedVector.push_back(node);
    }

    std::reverse(correctedVector.begin(), correctedVector.end());

    nodeVector = correctedVector;
}

static std::shared_ptr<Tensor> getWeightTensor(std::shared_ptr<Node> node)
{
    return std::static_pointer_cast<OperatorTensor>(node->getOperator())->getInput(1);
}

static std::shared_ptr<Tensor> getBiasTensor(std::shared_ptr<Node> node)
{
    return std::static_pointer_cast<OperatorTensor>(node->getOperator())->getInput(2);
}

std::vector<std::shared_ptr<Node>> retrieveNodeVector(std::shared_ptr<GraphView> graphView, bool newSchedule, bool verbose)
{
    std::vector<std::shared_ptr<Node>> nodeVector = graphView->getOrderedNodes();

    fixScheduling(nodeVector);
    nodeVector = removeMatchingNodes(nodeVector, "Producer");
    nodeVector = removeProdScalingNodes(nodeVector);

    if (verbose) 
    {
        Log::info("NB OF NODES = {}", nodeVector.size());
        for (std::shared_ptr<Node> node : nodeVector)
            Log::info("{} {}", node->type(), node->name());
    }

    return nodeVector;    
}

static std::shared_ptr<Node> getFirstNode(std::shared_ptr<GraphView> graphView)
{
    return retrieveNodeVector(graphView)[0];
}

// TODO : enhance this by modifying OperatorImpl in "core" ...
static DataType getDataType(std::shared_ptr<Node> node)
{
    auto op = std::static_pointer_cast<OperatorTensor>(node->getOperator());
    return op->getOutput(0)->dataType();
}

static std::shared_ptr<Aidge::Node> createScalingNode(std::string name, std::vector<std::string> attributes, double value)
{
    std::shared_ptr<Node> scalingNode = Mul(name);
  
    for (std::string attr : attributes)
        scalingNode->attributes()->addAttr("quantization.ptq." + attr, 0.0);
    
    // Add the scaling factor as a producer of the node

    std::shared_ptr<Tensor> scalingFactorTensor = std::make_shared<Tensor>(Array1D<double, 1> {value});
    std::shared_ptr<Node> scalingFactorProducer = addProducer(scalingNode, 1, {1}, "ScalingFactor"); 

    scalingFactorProducer->getOperator()->setOutput(0, scalingFactorTensor);

    // XXX graphView->add(scalingNode);

    return scalingNode;
}

bool insertScalingBelowProducer(std::shared_ptr<Node> producerNode, double scalingFactor, std::shared_ptr<GraphView> graphView)
{
    if (producerNode->attributes()->hasAttr("quantization.ptq.isProducerRounding"))
    {
        // In this case we 'bump' the node to the one above him (an actual ProducerScaling)
        // because the round node is not usable (only used when SSA is enabled)
        producerNode = producerNode->getParent(0);
    }

    if (producerNode->attributes()->hasAttr("quantization.ptq.isProducerScaling"))
    {
        // We accumulate the previous scaling factors by multiplying the SF of the ProducerScaling node 
        // (adding new nodes each time would make the graph unusable)
        multiplyScalingFactor(producerNode, scalingFactor);
        return true;
    }

    AIDGE_ASSERT(producerNode->type() == "Producer", " Cannot apply a scaling factor on node of type: {} which is not a Producer", producerNode->type());
   
    std::string scalingNodeName = makeUniqueName(producerNode->name() + "_ProducerScaling", graphView);
    std::shared_ptr<Node> scalingNode = createScalingNode(scalingNodeName, {"isProducerScaling"}, scalingFactor);

    scalingNode->getOperator()->setDataType(DataType::Float64); // getDataType(parentNode)
    scalingNode->getOperator()->setBackend(determineBackend(producerNode));

    insertChildren(producerNode, scalingNode, graphView);
    graphView->add(scalingNode->getParent(1)); // add the scaling factor producer

    return true;
}

// XXX HERE : Branches containing only Seamless nodes should be considered as residual too !!!
void insertResidualScalingNodes(std::shared_ptr<GraphView> graphView)
{
    // TODO: double check this ...

    std::vector<std::shared_ptr<Node>> nodeVector = retrieveNodeVector(graphView);

    for (std::shared_ptr<Node> node : nodeVector)
    {
        if (isMerging(node))
        {
            int nbParents = node->getParents().size();
            for (int i = 0; i < nbParents; i++)
            {
                std::shared_ptr<Node> parentNode = node->getParent(i);
                bool parentIsForking = (parentNode->getChildren().size() > 1);

                if (parentIsForking)
                {
                    // temporary verbose ...

                    Log::info(" ### found residual branch at index {}", i);
                    Log::info(" ### inserting multiplicative node ...");

                    std::string residualNodeName = makeUniqueName(parentNode->name() + "_Res", graphView);
                    std::shared_ptr<Node> residualNode = createScalingNode(residualNodeName, {"isScaling", "isResidual"}, 1.0);

                    residualNode->getOperator()->setDataType(DataType::Float64); // getDataType(parentNode)
                    residualNode->getOperator()->setBackend(determineBackend(parentNode));

                    graphView->insertParent(node, residualNode, i, 0, 0);
                    graphView->add(residualNode->getParent(1)); // add the scaling factor producer

                }
            }
        }
    }
}

static std::shared_ptr<Node> getPreviousScalingNode(std::shared_ptr<Node> node)
{
    std::shared_ptr<Node> currNode = node;
    while(!currNode->attributes()->hasAttr("quantization.ptq.isScaling"))
    {
        if (currNode->getParents().size() == 0)
        {
            Log::warn(" Warning : No previous Scaling node were found ! ");
            break;
        }
        currNode = currNode->getParents()[0];
    }
    return currNode;
}

void insertScalingNodes(std::shared_ptr<GraphView> graphView)
{
    insertResidualScalingNodes(graphView);

    std::set<std::shared_ptr<Node>> nodeSet = graphView->getNodes();

    for (std::shared_ptr<Node> parentNode : nodeSet)
    {
        if (isAffine(parentNode) || isMerging(parentNode) || isNotQuantized(parentNode))
        {
            std::string scalingNodeName = makeUniqueName(parentNode->name() + "_Scaling", graphView);
            std::shared_ptr<Node> scalingNode = createScalingNode(scalingNodeName, {"isScaling"}, 1.0);

            scalingNode->getOperator()->setDataType(DataType::Float64); // getDataType(parentNode)
            scalingNode->getOperator()->setBackend(determineBackend(parentNode));

            if (parentNode->getChildren().size() > 0) {
                insertChildren(parentNode, scalingNode, graphView);
            } else {
                // Log::info(" last node reached ! ");
                parentNode->addChild(scalingNode, 0, 0);
                graphView->add(scalingNode);
            }

            graphView->add(scalingNode->getParent(1)); // add the scaling factor producer

            // In the case the node is a non-linear operator we want to add an extra
            // scaling node before it to rescale it's input ...

            if (isNotQuantized(parentNode))
            {
                std::string prevScalingNodeName = makeUniqueName(parentNode->name() + "_PrevScaling", graphView);
                std::shared_ptr<Node> prevScalingNode = createScalingNode(prevScalingNodeName, {"isScaling"}, 1.0);

                prevScalingNode->getOperator()->setDataType(DataType::Float64); // getDataType(parentNode)
                prevScalingNode->getOperator()->setBackend(determineBackend(parentNode));

                graphView->insertParent(parentNode, prevScalingNode, 0, 0, 0);
                graphView->add(prevScalingNode->getParent(1)); // add the scaling factor producer
            }
        }
    }
}

// XXX double check this !
static bool nodeHasBias(std::shared_ptr<Node> node)
{
    if (node->getParents().size() == 3) {
        std::shared_ptr<Tensor> biasTensor = getBiasTensor(node);
        if (biasTensor)
            return true;
    }
    return false;
}

void normalizeParameters(std::shared_ptr<GraphView> graphView)
{
    // CREATE THE ACCUMULATED RATIO MAP ///////////////////////////////////////

    std::vector<std::shared_ptr<Node>> nodeVector = retrieveNodeVector(graphView);

    std::unordered_map<std::shared_ptr<Node>, double> accumulatedRatios;
    for (std::shared_ptr<Node> node : nodeVector)
        accumulatedRatios.insert(std::make_pair(node, 1.0));

    // ITERATE OVER THE GRAPH /////////////////////////////////////////////////

    std::shared_ptr<Node> firstNode = getFirstNode(graphView);

    for (std::shared_ptr<Node> node : nodeVector)
    {
        // Scaling nodes still have a ratio of 1, so they are seamless ...
        if (node->type() == "ReLU" || node->attributes()->hasAttr("quantization.ptq.isScaling") || isSeamless(node))
        {
            if (node != firstNode)
            {
                std::shared_ptr<Node> prevNode = node->getParent(0);
                accumulatedRatios[node] = accumulatedRatios[prevNode];
            }
        }

        // Residual nodes should enter in this category but their ratio is 1 ...
        if (isAffine(node))
        {
            // Rescale the weight tensor
            
            std::shared_ptr<Tensor> weightTensor = getWeightTensor(node);
            double scaling = getTensorAbsoluteMax(weightTensor);
            double ratio = 1.0 / scaling;

            //rescaleTensor(weightTensor, ratio);
            insertScalingBelowProducer(node->getParent(1), ratio, graphView);

            // Accumulate the ratio

            if (node == firstNode) {
                accumulatedRatios[node] = ratio;
            } else {
                std::shared_ptr<Node> prevNode = node->getParent(0);
                accumulatedRatios[node] = accumulatedRatios[prevNode] * ratio;
            }

            // Handle the bias ..

            if (nodeHasBias(node))
            {
                std::shared_ptr<Tensor> biasTensor = getBiasTensor(node);
                //rescaleTensor(biasTensor, accumulatedRatios[node] );
                insertScalingBelowProducer(node->getParent(2), accumulatedRatios[node], graphView);
            }
        }

        if (isNotQuantized(node))
        {
            // Gather the previous scaling factor

            std::shared_ptr<Node> prevScalingNode = getPreviousScalingNode(node);
            double prevRatio = accumulatedRatios[prevScalingNode];

            // Cancel the accumulated ratio

            multiplyScalingFactor(prevScalingNode, 1 / prevRatio);

            // Revert the canceling by using the next scaling node

            accumulatedRatios[node] = prevRatio;
            std::shared_ptr<Node> nextScalingNode = getUniqueChildren(node);
            multiplyScalingFactor(nextScalingNode, prevRatio);
        }

        if (isMerging(node))
        {
            std::vector<std::shared_ptr<Node>> mergingNodes = node->getParents();

            // Compute the max ratio ...

            double maxRatio = 0;
            for (std::shared_ptr<Node> mergingNode : mergingNodes)
            {
                double merginNodeRatio = accumulatedRatios[mergingNode];
                if (merginNodeRatio > maxRatio)
                    maxRatio = merginNodeRatio;
            }

            accumulatedRatios[node] = maxRatio;

            // Rescale the previous scaling Nodes
            for (std::shared_ptr<Node> mergingNode : mergingNodes)
            {
                double mergingNodeRatio = accumulatedRatios[mergingNode];
                double rescaling = mergingNodeRatio / maxRatio;

                std::shared_ptr<Node> scalingNode = getPreviousScalingNode(mergingNode);

                multiplyScalingFactor(scalingNode, 1 / rescaling);

                accumulatedRatios[mergingNode] /= rescaling; // optional ...
            }
        }
    }
}

// XXX TODO : take care of the CUDA backend for this too !!!
std::unordered_map<std::shared_ptr<Node>, double> computeRanges(std::shared_ptr<GraphView> graphView, std::shared_ptr<Tensor> inputTensor, bool scalingNodesOnly)
{
    std::unordered_map<std::shared_ptr<Node>, double> valueRanges;

    SequentialScheduler scheduler(graphView);
    scheduler.resetScheduling();

    // Inference ... 

    scheduler.forward(true, {inputTensor});

    // Gather ranges ...

    std::set<std::shared_ptr<Node>> nodeSet = graphView->getNodes();
    for (std::shared_ptr<Node> node : nodeSet)
    {
        if ((scalingNodesOnly && (node->attributes()->hasAttr("quantization.ptq.isScaling"))) || (!scalingNodesOnly && (node->type() != "Producer")))
        {
            std::shared_ptr<Operator> nodeOperator = node->getOperator();
            std::shared_ptr<Tensor> valueTensor = std::static_pointer_cast<Tensor> (nodeOperator->getRawOutput(0));
            double range = getTensorAbsoluteMax(valueTensor);

            // Associate the value to the scaling node ...
            valueRanges.insert(std::make_pair(node, range));
        }
    }

    return valueRanges;
}

std::unordered_map<std::shared_ptr<Node>, double> computeRanges(std::shared_ptr<GraphView> graphView, std::vector<std::shared_ptr<Tensor>> inputDataSet, bool scalingNodesOnly, bool useCuda)
{
    std::unordered_map<std::shared_ptr<Node>, double> valueRanges;
    std::set<std::shared_ptr<Node>> nodeSet = graphView->getNodes();
    
    // std::shared_ptr<Node> inputNode = getFirstNode(graphView);

    for (std::shared_ptr<Node> node : nodeSet)
        if ((scalingNodesOnly && (node->attributes()->hasAttr("quantization.ptq.isScaling"))) || (!scalingNodesOnly && (node->type() != "Producer")))
            valueRanges.insert(std::make_pair(node, 0));

    if (useCuda)
        graphView->setBackend("cuda");

    SequentialScheduler scheduler(graphView);
    scheduler.resetScheduling();

    int it = 0;

    for (std::shared_ptr<Tensor> sample : inputDataSet)
    {
        //Log::info(" IT : {}", it++);

        // Inference ...

        if (useCuda)
            sample->setBackend("cuda");

        scheduler.forward(true, {sample});

        // Gather the sample ranges ...

        std::unordered_map<std::shared_ptr<Node>, double> sampleRanges;
        for (std::shared_ptr<Node> node : nodeSet)
        {
            if ((scalingNodesOnly && (node->attributes()->hasAttr("quantization.ptq.isScaling"))) || (!scalingNodesOnly && (node->type() != "Producer")))
            {
                std::shared_ptr<Operator> nodeOperator = node->getOperator();
                std::shared_ptr<Tensor> valueTensor = std::static_pointer_cast<Tensor> (nodeOperator->getRawOutput(0));
                
                if (useCuda)
                    valueTensor->setBackend("cpu");

                double range = getTensorAbsoluteMax(valueTensor);

                // Associate the value to the scaling node ...
                sampleRanges.insert(std::make_pair(node, range));

                if (useCuda)
                    valueTensor->setBackend("cuda");
            }
        }

        // Update the global value ranges ...

        for (std::shared_ptr<Node> node : nodeSet)
        {
            if ((scalingNodesOnly && (node->attributes()->hasAttr("quantization.ptq.isScaling"))) || (!scalingNodesOnly && (node->type() != "Producer")))
                if (sampleRanges[node] > valueRanges[node])
                    valueRanges[node] = sampleRanges[node];
        }

        if (useCuda)
            sample->setBackend("cpu");
    }   

    if (useCuda)
        graphView->setBackend("cpu");

    return valueRanges;
}

void normalizeActivations(std::shared_ptr<GraphView> graphView, std::unordered_map<std::shared_ptr<Node>, double> valueRanges)
{
    std::shared_ptr<Node> firstNode = getFirstNode(graphView);

    // CREATE THE ACCUMULATED RATIO MAP ///////////////////////////////////////

    std::vector<std::shared_ptr<Node>> nodeVector = retrieveNodeVector(graphView);

    std::unordered_map<std::shared_ptr<Node>, double> accumulatedRatios;

    for (std::shared_ptr<Node> node : nodeVector)
        accumulatedRatios.insert(std::make_pair(node, 1.0));

    // ITERATE OVER THE GRAPH /////////////////////////////////////////////////

    for (std::shared_ptr<Node> node : nodeVector)
    {
        // Seamless scaling factor propagation ...

        if (isAffine(node) || isSeamless(node) || node->type() == "ReLU") 
        {
            if (node == firstNode) {
                accumulatedRatios[node] = 1.0;
            } else {
                std::shared_ptr<Node> prevNode = node->getParent(0);
                accumulatedRatios[node] = accumulatedRatios[prevNode];
            }
        }

        // Use the Scaling nodes to rescale the ranges ...

        if (node->attributes()->hasAttr("quantization.ptq.isScaling")) 
        {
            std::shared_ptr<Node> prevNode = node->getParent(0);

            double prevRatio = accumulatedRatios[prevNode];
            double nodeRange = valueRanges[node]; 

            multiplyScalingFactor(node, prevRatio / nodeRange);

            accumulatedRatios[node] = nodeRange;

            // If prevNode is Affine, fix the bias ...

            if (isAffine(prevNode)) {
                if (nodeHasBias(prevNode)) {                
                    insertScalingBelowProducer(prevNode->getParent(2), 1.0 / prevRatio, graphView);
                }
            }
        }

        // Merging nodes handling : use a maximum arbritration ...
        
        if (isMerging(node))
        {
            std::vector<std::shared_ptr<Node>> mergingNodes = node->getParents();

            // Compute the max ratio ...

            double maxRatio = 0;
            for (std::shared_ptr<Node> mergingNode : mergingNodes)
            {
                double mergingNodeRatio = accumulatedRatios[mergingNode];
                if (mergingNodeRatio > maxRatio)
                    maxRatio = mergingNodeRatio;
            }

            accumulatedRatios[node] = maxRatio;

            for (std::shared_ptr<Node> mergingNode : mergingNodes)
            {
                double mergingNodeRatio = accumulatedRatios[mergingNode];
                std::shared_ptr<Node> scalingNode = getPreviousScalingNode(mergingNode);
                multiplyScalingFactor(scalingNode, mergingNodeRatio / maxRatio);
                // Log::notice(" SCALING NODE : {} {}", scalingNode->type(), scalingNode->name());
            }
        }

        if (isNotQuantized(node))
        {
            std::shared_ptr<Node> prevScalingNode = node->getParent(0);
            double prevRatio = accumulatedRatios[prevScalingNode];
            Log::notice(" prev ratio : {} ", prevRatio);

            // This causes the previous range to not full fill the [-1, 1] interval !!!
            // It could be avoided by systematicly add an extra Scaling node before each 
            // non linearity ...

            multiplyScalingFactor(prevScalingNode, prevRatio);
        }

    }
}

std::unordered_map<std::shared_ptr<Node>, std::pair<bool, bool>> computeSignMap(std::shared_ptr<GraphView> graphView, bool verbose)
{
    std::shared_ptr<Node> firstNode = getFirstNode(graphView);

    std::unordered_map<std::shared_ptr<Node>, std::pair<bool, bool>> signMap;

    std::pair<bool, bool> unsignedPair(true, true);
    for (std::shared_ptr<Node> node : graphView->getNodes())
        if (node->type() != "Producer")
            signMap.insert(std::make_pair(node, unsignedPair));

    // ITERATE OVER THE GRAPH

    std::vector<std::shared_ptr<Node>> nodeVector = retrieveNodeVector(graphView);

    for (std::shared_ptr<Node> node : nodeVector)
    {
        bool isFirstNode = (node == firstNode);

        if (isAffine(node))
        {   
            // Affine nodes always have a single parent 
            if (!isFirstNode) 
                signMap[node].first = signMap[node->getParent(0)].second;
            else
                signMap[node].first = false;

            signMap[node].second = false;
        } 

        if (node->attributes()->hasAttr("quantization.ptq.isScaling")) 
        {
            signMap[node].second = false;

            // Scaling nodes always have a single parent 
            std::shared_ptr<Node> parent = node->getParent(0); 

            bool allChildrenAreReLU = true;
            allChildrenAreReLU &= !(node->getChildren().empty()); // a bit convoluted ...
            for (std::shared_ptr <Node> child : node->getChildren())
                allChildrenAreReLU &= (child->type() == "ReLU");

            // Correct the previous single node (when it is an Affine node) ...
            if (allChildrenAreReLU)
                if (isAffine(parent) || isMerging(parent))
                    signMap[parent].second = true;

            // Maintain unsigned output
            if (signMap[parent].second)
                signMap[node].second = true;

            // Set the link ...
            signMap[node].first = signMap[parent].second;
        }
       
        if (isMerging(node))
        {
            std::vector<std::shared_ptr<Node>> parentNodes = node->getParents(); 

            bool allParentAreSigned = true;
            bool allParentAreUnsigned = true;
            for(std::shared_ptr<Node> parent : parentNodes)
            {
                bool parentSign = signMap[parent].second;
                allParentAreSigned &= !parentSign;
                allParentAreUnsigned &= parentSign;
            }

            if (allParentAreSigned)
                signMap[node] = std::make_pair(false, false);
            else if (allParentAreUnsigned)
                signMap[node] = std::make_pair(true, true);
            else 
            {
                // Arbitration : Signed type wins !
                for(std::shared_ptr<Node> parent : parentNodes)
                {
                    while (!parent->attributes()->hasAttr("quantization.ptq.isScaling"))
                    {
                        signMap[parent] = std::make_pair(false, false);
                        // We are on a branch so nodes always have 1 parent ...
                        parent = parent->getParent(0); 
                    }

                    signMap[parent].second = false;
                }

                signMap[node].first = false;
            }
        }
        
        if (node->type() == "ReLU" || isSeamless(node) || isNotQuantized(node))
        {
            // Thoses nodes always have a single parent 
            std::shared_ptr<Node> parent = node->getParent(0); 
            if (parent)
            {
                signMap[node].first = signMap[parent].second;
                signMap[node].second = signMap[node].first;
            }
            
        }
    }

    // VERBOSE

    if (verbose)
    {
        Log::info(" === SIGN MAP === ");
        for (std::shared_ptr<Node> node : nodeVector)
            Log::info(" {}{} | {}", static_cast<int>(signMap[node].first), static_cast<int>(signMap[node].second), node->name());
    }

    // SANITY CHECK (TEMPORARY)

    for (std::shared_ptr<Node> node : nodeVector)
    {
        for (std::shared_ptr<Node> child : node->getChildren()) 
        {
            if (signMap[node].second != signMap[child].first)
                Log::error(" computeSignMap : link is not sane ! ({} -> {})", node->name(), child->name());                    
        }
    }
    
    return signMap;
}

void quantizeNormalizedNetwork(std::shared_ptr<GraphView> graphView, std::uint8_t nbBits, bool noQuant, bool optimizeSigns, bool verbose)
{
    if (optimizeSigns && noQuant)
        AIDGE_THROW_OR_ABORT(std::runtime_error, " Sign-optimization can not be applied if network is not fully quantized ...");

    double signedMax = (1 << (nbBits - 1)) - 1;
    double unsignedMax = (1 << nbBits) - 1;

    std::unordered_map<std::shared_ptr<Node>, std::pair<bool, bool>> signMap;

    if (optimizeSigns)
        signMap = computeSignMap(graphView, verbose);
    else
    {
        std::pair<bool, bool> signedPair(false, false);
        for (std::shared_ptr<Node> node : graphView->getNodes())
            if (node->type() != "Producer")
                signMap.insert(std::make_pair(node, signedPair));
    }

    // ITERATE OVER THE GRAPH /////////////////////////////////////////////////

    std::vector<std::shared_ptr<Node>> nodeVector = retrieveNodeVector(graphView);

    for (std::shared_ptr<Node> node : nodeVector)
    {
        if (isAffine(node))
        {
            // Rescale the weight tensor
            std::shared_ptr<Tensor> weightTensor = getWeightTensor(node);
            insertScalingBelowProducer(node->getParent(1),signedMax,graphView);

            if (!noQuant)
                insertRoundBelowProducer(node->getParent(1),graphView);

            // Rescale the bias tensor
            if (nodeHasBias(node))  
            {
                bool inputIsUnsigned = signMap[node].first;
                double rescaling = inputIsUnsigned ? unsignedMax * signedMax : signedMax * signedMax;
            
                std::shared_ptr<Tensor> biasTensor = getBiasTensor(node);
                insertScalingBelowProducer(node->getParent(2),rescaling,graphView);

                if (!noQuant)
                    insertRoundBelowProducer(node->getParent(2),graphView);
            }

            // Compensate the rescaling using the next Scaling node

            double rescaling = 1.0 / signedMax;

            bool inputIsUnsigned  = signMap[node].first; 
            bool outputIsUnsigned = signMap[node].second;

            rescaling /= inputIsUnsigned  ? unsignedMax : signedMax;
            rescaling *= outputIsUnsigned ? unsignedMax : signedMax;
            
            std::shared_ptr<Node> scalingNode = getUniqueChildren(node); // TODO : assert if scalingNode is a Scaling ...

            multiplyScalingFactor(scalingNode,rescaling) ;          
        }
        
        if (isMerging(node))
        {
            double rescaling = 1.0;

            bool inputIsUnsigned  = signMap[node].first;
            bool outputIsUnsigned = signMap[node].second;

            rescaling /= inputIsUnsigned  ? unsignedMax : signedMax;
            rescaling *= outputIsUnsigned ? unsignedMax : signedMax;

            std::shared_ptr<Node> scalingNode = getUniqueChildren(node); // TODO : assert if scalingNode is a Scaling ...
        
            multiplyScalingFactor(scalingNode, rescaling) ;          
        }

        if (isNotQuantized(node))
        {
            double rescaling = 1 / signedMax;  // XXX handle the signs !!!

            std::shared_ptr<Node> prevScalingNode = node->getParent(0);
            multiplyScalingFactor(prevScalingNode, rescaling);

            std::shared_ptr<Node> nextScalingNode = getUniqueChildren(node);
            multiplyScalingFactor(nextScalingNode, 1 / rescaling);
        }
        
        // Handle the Scaling Nodes ...

        if (node->attributes()->hasAttr("quantization.ptq.isScaling"))
        {
            // Don't touch the scalings that precede non-linearities ...

            bool precedesNonLinearNode = false;
            if (node->getChildren().size() == 1)
                if (isNotQuantized(getUniqueChildren(node)))
                    precedesNonLinearNode = true; 

            if (!noQuant && !precedesNonLinearNode) 
            {  
                // Replace the Scaling Node by a Quantizer

                auto scalingFactorTensor = std::static_pointer_cast<OperatorTensor>(node->getOperator())->getInput(1);
                std::shared_ptr<Tensor> fallback;
                const Tensor& localTensor = scalingFactorTensor->refCastFrom(fallback, DataType::Float64, "cpu");
                double oldScalingFactor = localTensor.get<double>(0); //!\\ 

                std::shared_ptr<Node> quantizerNode = Quantizer(oldScalingFactor, -(signedMax + 1), signedMax, node->name());
                quantizerNode->getOperator()->setDataType(DataType::Float64); // getDataType(parentNode)
                quantizerNode->getOperator()->setBackend(determineBackend(node));
                graphView->replace({node, node->getParent(1)}, {quantizerNode});

                if (optimizeSigns)
                {
                    double rescaling = 1.0;

                    bool inputIsUnsigned  = signMap[node].first;
                    bool outputIsUnsigned = signMap[node].second;

                    rescaling /= inputIsUnsigned  ? unsignedMax : signedMax;
                    rescaling *= outputIsUnsigned ? unsignedMax : signedMax;

                    double currScalingFactor = getScalingFactor(quantizerNode);
                    updateScalingFactor(quantizerNode, currScalingFactor * rescaling);
                    
                    if(outputIsUnsigned)
                        setClipRange(quantizerNode, 0, unsignedMax);                 
                }
            }
        }
    }
}

static void insertCompensationNodes(std::shared_ptr<GraphView> graphView, std::uint8_t nbBits)
{
    // XXX Use the signMap to increase the resolution when possible ...
    double signedMax = (1 << (nbBits - 1)) - 1;    

    std::vector<std::shared_ptr<Node>> nodeVector = retrieveNodeVector(graphView);

    for (std::shared_ptr<Node> node : nodeVector)
    {
        // The appropriate strategy is to check if the Quantizer is not   
        // preceded by an Weighted node (that is not forking), and insert  
        // a coeff node (Compensation) if so ...
 
        if (node->type() == "Quantizer")
        {
            // Note : this works because a Quantizer has only one Parent ...

            std::shared_ptr<Node> parentNode = node->getParent(0);
            bool parentHasWeight = isAffine(parentNode);
            bool parentIsForking = (parentNode->getChildren().size() > 1);

            if (parentIsForking || !parentHasWeight) // insert a Compensation Node ...
            {   
                // Create and insert the multiplicative node before the Quantizer

                std::string mulNodeName = makeUniqueName(node->name() + "_Mul", graphView);
                std::shared_ptr<Node> mulNode = Mul(mulNodeName);
                
                mulNode->attributes()->addAttr("quantization.ptq.isCompensation", 0.0);
                mulNode->getOperator()->setDataType(DataType::Float64); // getDataType(parentNode)
                mulNode->getOperator()->setBackend(determineBackend(node));

                graphView->insertParent(node, mulNode, 0, 0, 0);

                // Add the coeff producer to the multiplier node

                std::shared_ptr<Node> coeffProducer = addProducer(mulNode, 1, {1}, ""); 
                std::shared_ptr<Tensor> coeffTensor = std::make_shared<Tensor>(Array1D<double, 1> {signedMax});
                coeffProducer->getOperator()->setOutput(0, coeffTensor);

                coeffProducer->getOperator()->setDataType(DataType::Float64);
                coeffProducer->getOperator()->setBackend(determineBackend(node)); 

                graphView->add(coeffProducer); // needed ?

                // Adapt the scaling factor value accordingly

                double currScalingFactor = getScalingFactor(node); 
                updateScalingFactor(node, currScalingFactor / signedMax); 
            }
        }
    }
}

void performSingleShiftApproximation(std::shared_ptr<GraphView> graphView, bool noQuant)
{
    std::vector<std::shared_ptr<Node>> nodeVector = retrieveNodeVector(graphView);

    for (std::shared_ptr<Node> node : nodeVector)
    {
        if (node->type() == "Quantizer")
        {
            std::shared_ptr<Node> linearNode = node->getParent(0);

            double base = getScalingFactor(node);

            double approx = std::pow(2, std::ceil(std::log2(base)));

            updateScalingFactor(node, approx);

            double ratio = base / approx;

            insertScalingBelowProducer(linearNode->getParent(1), ratio, graphView);
            if (!noQuant)
                insertRoundBelowProducer(linearNode->getParent(1), graphView);

            if (nodeHasBias(linearNode))
            {
                insertScalingBelowProducer(linearNode->getParent(2), ratio, graphView);
                if (!noQuant)
                    insertRoundBelowProducer(linearNode->getParent(2), graphView);
            }
        }
    }
}


static void printScalingFactors(std::shared_ptr<GraphView> graphView)
{
    for (auto node : retrieveNodeVector(graphView))
        if (node->attributes()->hasAttr("quantization.ptq.isScaling") || node->type() == "Quantizer")
        {
            double scalingFactor = getScalingFactor(node);
            Log::info(" {:.6f} ({})", scalingFactor, node->name());
        }
}

static void setupDataType(std::shared_ptr<GraphView> graphView, std::vector<std::shared_ptr<Tensor>> inputDataSet, DataType dataType)
{
    graphView->setDataType(dataType);

    for (auto inputNode : graphView->inputNodes()) {
        auto op = std::static_pointer_cast<OperatorTensor>(inputNode->getOperator());
        auto inputTensor = op->getInput(0);
        if (inputTensor)
            inputTensor->setDataType(dataType);
    }

    for (auto tensor : inputDataSet)
        tensor->setDataType(dataType);
}

void quantizeNetwork(std::shared_ptr<GraphView> graphView, std::uint8_t nbBits, std::vector<std::shared_ptr<Tensor>> inputDataSet, Clipping clippingMode, bool noQuant, bool optimizeSigns, bool singleShift, bool useCuda, bool verbose)
{
    Log::notice(" === QUANT PTQ 0.2.21 === ");

    graphView->setBackend("cpu");

    if (!checkArchitecture(graphView))
        return;

    DataType initialDataType = (inputDataSet[0])->dataType();
    setupDataType(graphView, inputDataSet, DataType::Float64);

    Log::notice(" Preparing the network for the PTQ ... ");
    prepareNetwork(graphView);

    Log::notice(" Inserting the scaling nodes ...");
    insertScalingNodes(graphView);

    crossLayerEqualization(graphView);

    Log::notice(" Normalizing the parameters ...");
    normalizeParameters(graphView);

    Log::notice(" Computing the value ranges ...");
    std::unordered_map<std::shared_ptr<Node>, double> valueRanges = computeRanges(graphView, inputDataSet, true, useCuda);

    //Log::info(" === RANGES (BEFORE ADJUST) ===");

    Log::notice(" Optimizing the clipping values ...");
    valueRanges = adjustRanges(clippingMode, valueRanges, nbBits, graphView, inputDataSet, useCuda, verbose);

    //Log:debug("=== RANGES (AFTER ADJUST) ===");
    //printRanges(graphView, valueRanges);
    Log::notice(" Normalizing the activations ...");
    normalizeActivations(graphView, valueRanges);

    Log::notice(" Quantizing the normalized network ...");
    quantizeNormalizedNetwork(graphView, nbBits, noQuant, optimizeSigns, verbose);

    if (singleShift)
    {
        Log::notice( " Inserting the compensation nodes ...");
        insertCompensationNodes(graphView, nbBits);

        Log::notice(" Performing the Single-Shift approximation ...");
        performSingleShiftApproximation(graphView, noQuant);
    }

    if (verbose)
        printScalingFactors(graphView);

    if (useCuda)
        graphView->setBackend("cuda");

    Log::notice(" Reseting the scheduler ...");
    SequentialScheduler scheduler(graphView);
    scheduler.resetScheduling();

    Log::notice(" Network is quantized !");
}

std::unordered_map<std::string, double> getWeightRanges(std::shared_ptr<GraphView> graphView)
{
    std::unordered_map<std::string, double> weightRanges;

    for (std::shared_ptr<Node> node : graphView->getNodes())
    {
        if (isAffine(node))
        {
            std::shared_ptr<Tensor> weightTensor = getWeightTensor(node);
            double range = getTensorAbsoluteMax(weightTensor);
            weightRanges.insert(std::make_pair(node->name(), range));
        }
    }

    return weightRanges;
}

void clearBiases(std::shared_ptr<GraphView> graphView) 
{
    for (std::shared_ptr<Node> node : graphView->getNodes()) {
        if (node->type() == "FC" || node->type() == "Conv2D") {
            std::shared_ptr<Tensor> biasTensor = std::static_pointer_cast<OperatorTensor>(node->getOperator())->getInput(2);
            //rescaleTensor(biasTensor, 0);
            insertScalingBelowProducer(node->getParent(2),0,graphView);
        }
    }
}
}
